"""Utilities the connect to Stefan Duchesne potatopeel module, which is responsible for
peeling sources in racs. The repository is available at:

https://gitlab.com/Sunmish/potato/-/tree/main

Potato stands for "Peel Out That Annoying Terrible Object".

Although this is a python module, for the moment it is expected
to be in a singularity container. There are several reasons
for this, but the principal one is that the numba module used
by potatopeel may be difficult to get working correctly alongside
dask and flint. Keeping it simple at this point is the main aim.
There is also the problem of casatasks + python-casacore not
jiving in newer python versions.

"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Collection, NamedTuple

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table
from casacore.tables import table

from flint.imager.wsclean import WSCleanOptions
from flint.logging import logger
from flint.ms import get_freqs_from_ms, get_phase_dir_from_ms
from flint.naming import get_potato_output_base_path
from flint.options import MS
from flint.sclient import run_singularity_command
from flint.utils import (
    create_directory,
    generate_strict_stub_wcs_header,
    get_packaged_resource_path,
)


class PotatoConfigOptions(NamedTuple):
    """Container class to hold options that go into the potatopy
    configuration creation software by Stefan Duchesne. See:

    https://gitlab.com/Sunmish/potato
    """

    image_size: int = 6148
    """Size of an in-field image"""
    image_scale: float = 0.0006944
    """The pixel scale of the in-field image in degrees"""
    image_briggs: float = -1.5
    """Briggs robust parameter for the in-field image"""
    image_channels: int = 4
    """Number of output channels for the in-field image"""
    image_minuvl: float = 700
    """"Minimum (u,v)- distance in wavelengths for data to be selected"""
    peel_size: int = 1000
    """Size of the peel image to make, in pixels"""
    peel_scale: float = 0.0006944
    """Pixel scale of the peel images in degrees"""
    peel_channels: int = 16
    """Number of output channels for the peel images"""
    peel_nmiter: int = 7
    """Number of major iterations allowed for the peel sources"""
    peel_minuvl: float = 700
    """"Minimum (u,v)- distance in wavelengths for data to be selected for the peel image"""
    peel_multiscale: bool = True
    """Whether multi-scale is to be used for the peel sources"""


class PotatoPeelArguments(NamedTuple):
    """The mandatory arguments for potato peel."""

    # These are kept separatte from the PotatoPeelOptions
    # class so the PotatoPeelOptions may be written to a template file
    ms: Path
    """The measurement set that will be examined for peeling"""
    ras: Collection[float]
    """The source RA in degrees to peel"""
    decs: Collection[float]
    """The source Dec in degrees to peel"""
    peel_fovs: Collection[float]
    """The field-of-views that should be created for the peel source in degrees"""
    image_fov: float
    """The field-of-view in degrees of the main in-field image. If a sources is within this radius it is not peeled (because it would be imaged)"""
    n: Collection[str]
    """Name of the source being peeled"""


class PotatoPeelOptions(NamedTuple):
    """Container class to hold options that go to the potato peel
    software by Stefan Duchesne. FLINT uses the `hot_potato` version.
    ee:

    https://gitlab.com/Sunmish/potato
    """

    c: Path | None = None
    """Path to the potatopeel configuration file"""
    solint: float = 30
    """Solution interval to use when applying gaincal"""
    calmode: str = "P"
    """Self-calibration mode to use (see casatasks gaincal)"""
    minpeelflux: float = 0.5
    """Minimum flux, in Jy, for the peeling procedure (image->selfcal->image)"""
    refant: int = 1
    """Reference antenna to use when solving for self-cal solutions"""
    direct_subtract: bool = True
    """Whether a direct model subtraction (without self-cal) should be used ift he source is faint"""
    intermediate_peels: bool = True
    """Creates an image after each calibration and subtraction loop to show iterative improvements of the subject peel source"""
    T: str | Path = "peel"
    """Where the temporary wsclean files will be written to"""
    minuvimage: float | None = None
    """The minimum uv distance in wavelengths to use for imaging"""
    minuvpeel: float | None = None
    """The minimum uv distance in wavelengths to use when attempting to self-calibrate"""

    def with_options(self, **kwargs) -> PotatoPeelOptions:
        items = self._asdict()
        items.update(**kwargs)
        return PotatoPeelOptions(**items)


def load_known_peel_sources() -> Table:
    """Locate and load the packaged set of known sources to peel. These sources
    are drawn from Duchesne et. al. (2023), Table 3.

    Returns:
        Table: The astropy Table of candidate sources to remove.
    """
    peel_srcs_csv = get_packaged_resource_path("flint", "data/peel/known_sources.csv")

    logger.info(f"Loading in {peel_srcs_csv=}")
    peel_tab = Table.read(peel_srcs_csv)

    logger.info(f"Loaded {len(peel_tab)} sources. ")

    return peel_tab


def source_within_image_fov(
    source_coord: SkyCoord,
    beam_coord: SkyCoord,
    image_size: int,
    pixel_scale: u.Quantity | str,
) -> bool:
    """Evaluate whether a source will be within the field of view
    of an image.

    Args:
        source_coord (SkyCoord): The source position to consider
        beam_coord (SkyCoord): The center beam position to consider
        image_size (int): The image size. This assumes a square image.
        pixel_scale (Union[u.Quantity, str]): The pixel size, assuming square pixels.

    Returns:
        bool: Whether the source is expected to be in the field of view
    """
    # TODO: Allow the wcs to be passed in

    wcs = generate_strict_stub_wcs_header(
        position_at_image_center=beam_coord,
        image_shape=(image_size, image_size),
        pixel_scale=pixel_scale,
    )

    x, y = wcs.world_to_pixel(source_coord)

    # Since the reference pixel is at the image center, then the valid domain
    # is between 0 and image size
    source_in_image = 0 <= x < image_size and 0 <= y < image_size

    return source_in_image


def find_sources_to_peel(
    ms: MS,
    image_options: WSCleanOptions,
    field_idx: int = 0,
    maximum_offset: float = 30,
    minimum_apparent_brightness: float = 0.5,
    override_beam_position_with: SkyCoord | None = None,
) -> Table | None:
    """Obtain a set of sources to peel from a reference candidate set. This will
    evaluate whether a source should be peels based on two criteria:

    - if it is below a nominal primary beam cut off level (10 percent, assuming a gaussian primary beam)
    - if the sources is within some separation of the imaging center

    Args:
        ms (MS): The measurement set that is being considered
        image_options (WSCleanOptions): The imaging parameters that will be used to compute a placeholder WCS
        field_idx (int, optional): Which field in the MS to draw the position from. Defaults to 0.
        maximum_offset (float, optional): The largest separation, in degrees, before a source is ignored. Defaults to 30.0.
        minimum_apparent_brightness (float, optional): The minimum apparent brightnessm, in Jy, a source should be before attempting to peel. Defaults to 0.5.
        override_beam_position_with (Optional[SkyCoord], optional): Ignore the beam position of the input MS, instead use this. Do not rely on this option as it may be taken away. Defaults to None.

    Returns:
        Union[Table,None]: Collection of sources to peel from the reference table. Column names are Name, RA, Dec, Aperture. This is the package table. If no sources need to be peeled None is returned.
    """
    image_size, pixel_scale = None, None

    if isinstance(image_options, WSCleanOptions):
        image_size = image_options.size
        pixel_scale = image_options.scale
        logger.info("Replace known wsclean units with astropy.unit.Quantity version")
        pixel_scale = pixel_scale.replace("asec", "arcsec")
        pixel_scale = pixel_scale.replace("amin", "arcmin")
    else:
        raise TypeError(f"{type(image_options)=} is not known. ")

    logger.debug(f"Extracting image direction for {field_idx=}")
    image_coord = (
        get_phase_dir_from_ms(ms=ms)
        if override_beam_position_with is None
        else override_beam_position_with
    )

    logger.info(
        f"Considering sources to peel around {image_coord=}, {type(image_coord)=}"
    )

    peel_srcs_tab = load_known_peel_sources()

    freqs = get_freqs_from_ms(ms=ms)
    nominal_freq = np.mean(freqs) * u.Hz  # type: ignore
    logger.info(f"The nominal frequency is {nominal_freq.to(u.MHz)}")

    peel_srcs = []

    # TODO: Make this a mapping function
    for src in peel_srcs_tab:
        src_coord = SkyCoord(src["RA"], src["Dec"], unit=(u.degree, u.degree))
        offset = image_coord.separation(src_coord)
        if source_within_image_fov(
            source_coord=src_coord,
            beam_coord=image_coord,
            image_size=image_size,
            pixel_scale=pixel_scale,
        ):
            logger.debug(
                f"Source {src['Name']} within image field of view of image with {image_size}x{image_size} and {pixel_scale=}"
            )
            continue

        if offset > maximum_offset * u.deg:  # type: ignore
            continue

        logger.info(
            f"Source {src['Name']} is {offset} away, below {maximum_offset=} cutoff "
        )

        # At the moment no brightness is in the known_sources.csv
        # taper = generate_pb(
        #     pb_type="airy", freqs=nominal_freq, aperture=12 * u.m, offset=offset
        # )
        # assert isinstance(taper.atten, float), "Expected a float"

        # if taper.atten[0] * tab[] > cutoff:
        #     logger.info(
        #         f"{src['Name']} attenuation {taper.atten} is above {cutoff=} (in field of view)"
        #     )
        #     continue

        peel_srcs.append(src)

    return Table(rows=peel_srcs, names=peel_srcs_tab.colnames) if peel_srcs else None


def prepare_ms_for_potato(ms: MS) -> MS:
    """The potatopeel software requires the data column being operated against to be
    called DATA. This is a requirement of CASA and its gaincal / applysolution task.

    If there is already a DATA column and this is not the nominated data column described
    by the MS then it is removed, and the nominated data column is renamed.

    If the DATA column already exists and it is the nominated column described by
    MS then this is returned. If there is a CORRECTED_DATA column in this situation
    it is removed.

    Args:
        ms (MS): The measurement set that will be edited

    Raises:
        ValueError: Raised when a column name is expected but can not be found.

    Returns:
        MS: measurement set with the data column moved into the correct location
    """
    data_column = ms.column

    logger.info(f"The nominated column is: {data_column=}")
    logger.warning(
        "Deleting and renaming columns so final column is DATA. "
        "PotatoPeel only operates on the DATA column. "
    )

    # If the data column already exists and is the nominated column, then we should
    # just return, ya scally-wag
    if data_column == "DATA":
        logger.info(f"{data_column=} is already DATA. No need to rename. ")

        with table(str(ms.path), readonly=False, ack=False) as tab:
            if "CORRECTED_DATA" in tab.colnames():
                logger.info(f"Removing CORRECTED_DATA column from {ms.path}")
                tab.removecols("CORRECTED_DATA")

        return ms

    with table(str(ms.path), ack=False, readonly=False) as tab:
        colnames = tab.colnames()
        if data_column not in colnames:
            raise ValueError(
                f"Column {data_column} not found in {ms.path!s}. Columns found: {colnames}"
            )

        # In order to rename the data_column to DATA, we need to make sure that
        # there is not an existing DATA column
        if "DATA" in colnames:
            logger.info("Removing the existing DATA column")
            tab.removecols("DATA")

        logger.info(f"Renaming {data_column} to DATA")
        tab.renamecol(data_column, "DATA")

        # Update column names after the delete and rename
        colnames = tab.colnames()

        # Remove any CORRECT_DATA column, should it exist, as
        # potatopeel will create it
        if "CORRECTED_DATA" in colnames:
            logger.info(f"Removing CORRECTED_DATA column from {ms.path}")
            tab.removecols("CORRECTED_DATA")

    return ms.with_options(column="DATA")


def _potato_options_to_command(
    potato_options: PotatoPeelArguments | PotatoConfigOptions | PotatoPeelOptions,
    skip_keys: Collection[str] | None = None,
    check_double_keys: bool = False,
) -> str:
    """Construct the CLI options that would be provided to
    a potato peel CLI program

    Args:
        potato_options (Union[PotatoPeelArguments,PotatoConfigOptions, PotatoPeelOptions]): An instance of one of the option classes to draw from
        skip_keys (Optional[Collection[str]], optional): A collections of keys to ignore when build the CLI. If None all keys in the provided options instance are used. Defaults to None.
        check_double_leys (bool, optional): Some long form names in `hot_potato` are single dash while others are double dash. This is not the case in the config creation tool. This will check to see if the double should be used. Defaults to False.

    Raises:
        TypeError: When an unrecognised data type is found in the provided options class

    Returns:
        str: A string of the CLI options and keys
    """
    skip_keys = tuple(skip_keys) if skip_keys else tuple()

    DOUBLE = ("ras", "decs", "peel_fovs", "intermediate_peels", "direct_subtract")

    sub_options = ""
    for key, value in potato_options._asdict().items():
        flag = "--"
        if check_double_keys:
            flag = "--" if key in DOUBLE else "-"
        logger.debug(f"{key=} {value=} {type(value)=}")
        if key in skip_keys:
            logger.debug(f"{key=} in {skip_keys=}, skipping")
            continue
        if isinstance(value, bool):
            logger.debug("bool")
            if value:
                sub_options += f"{flag}{key} "
        elif isinstance(value, (tuple, list)):
            logger.debug("tuple or list")
            out_value = " ".join([f"{v}" for v in value])
            sub_options += f"{flag}{key} {out_value} "
        elif isinstance(value, (int, float, str)):
            logger.debug("int flot str")
            sub_options += f"{flag}{key} {value} "
        elif isinstance(value, Path):
            logger.debug("Path")
            sub_options += f"{flag}{key} {value!s} "
        elif value is None:
            continue
        else:
            raise TypeError(f"Unrecognised {key=} {value=} type")

    return sub_options


class PotatoConfigCommand(NamedTuple):
    """Container for potato configuration command results"""

    config_path: Path
    """Path to the configuration file generated"""
    command: str
    """The command string that should be executed"""


def _potato_config_command(
    config_path: Path, potato_config_options: PotatoConfigOptions
) -> PotatoConfigCommand:
    """Create the peel_configuration.py command that will be called
    in the potato singularity image. This is the CLI version of the
    code (not calling the python function).

    Args:
        config_path (Path): Output location of the configuration file
        potato_config_options (PotatoConfigOptions): Instance of all the options to use

    Returns:
        PotatoconfigCommand: The CLI command that will be executed to create a potato configuration file
    """

    command = f"peel_configuration.py {config_path!s} "

    sub_options = _potato_options_to_command(potato_options=potato_config_options)
    command = command + sub_options

    logger.debug(f"Constructed command {command}")
    return PotatoConfigCommand(config_path=config_path, command=command)


def create_run_potato_config(
    potato_container: Path,
    ms_path: Path | MS,
    potato_config_options: PotatoConfigOptions,
) -> PotatoConfigCommand:
    """Construct and run a CLI command into the `peel_configuration.py`
    script of the `potatopeel` package.

    Args:
        potato_container (Path): Container with the `potatopeel` package installed
        ms_path (Union[Path, MS]): Path to the measurement set that will be peeled
        potato_config_options (PotatoConfigOptions): Options to tweak the values in the peel configuration

    Returns:
        PotatoConfigCommand: Container of the path to the peel configuration file and the corresponding command that generated it
    """

    ms = MS.cast(ms=ms_path)
    base_potato_path = get_potato_output_base_path(ms_path=ms.path)
    config_path = base_potato_path.parent / (base_potato_path.name + ".config")
    potato_config_command = _potato_config_command(
        config_path=config_path, potato_config_options=potato_config_options
    )

    run_singularity_command(
        image=potato_container,
        command=potato_config_command.command,
        bind_dirs=potato_config_command.config_path.parent,
    )

    return potato_config_command


class PotatoPeelCommand(NamedTuple):
    """Container to hold the items of the hot potato command"""

    ms: MS
    """The measurement set that potato has operated against"""
    command: str
    """The hot potato command that will be executed"""


def _potato_peel_command(
    ms: MS,
    potato_peel_arguments: PotatoPeelArguments,
    potato_peel_options: PotatoPeelOptions,
) -> PotatoPeelCommand:
    """Construct the CLI command for `hot_potato`, and appropriately
    handle the mandatory and optional arguments.

    Args:
        ms (MS): The measurement set that will be peeled
        potato_peel_arguments (PotatoPeelArguments): The mandatory arguments for `hot_potato`
        potato_peel_options (PotatoPeelOptions): The `hot_potato` options to supply

    Returns:
        PotatoPeelCommand: The `hot_potato` command that was constructed from the input `PotatoPeelOptions`
    """

    command = (
        f"hot_potato {ms.path.absolute()!s} {potato_peel_arguments.image_fov:.4f} "
    )

    for item in [potato_peel_arguments, potato_peel_options]:
        # The skip keys handle the mandatory arguments that are specified above
        sub_options = _potato_options_to_command(
            potato_options=item,  # type: ignore
            skip_keys=("image_fov", "ms"),
            check_double_keys=True,
        )
        command += sub_options

    return PotatoPeelCommand(ms=ms, command=command)


def create_run_potato_peel(
    potato_container: Path,
    ms: MS,
    potato_peel_arguments: PotatoPeelArguments,
    potato_peel_options: PotatoPeelOptions,
) -> PotatoPeelCommand:
    """Construct and run a `hot_potato` command to peel out sources from
    a measurement set.

    Args:
        potato_container (Path): Container with the potato peel software and appropriate tools (including wsclean)
        ms (MS): The measurement set that contains sources to peel
        potato_peel_arguments (PotatoPeelArguments): The mandatory arguments that go into `hot_potato`.
        potato_peel_options (PotatoPeelOptions): Options that are supplied to `hot_potato`.

    Returns:
        PotatoPeelCommand: The executed `hot_potato` command
    """

    # Construct the command
    potato_peel_command = _potato_peel_command(
        ms=ms,
        potato_peel_arguments=potato_peel_arguments,
        potato_peel_options=potato_peel_options,
    )

    # make sure the container can bind to all necessary directories. This
    # includes the potential directory used by wsclean to temporarily store
    # files
    bind_dirs = [
        ms.path,
    ]
    if potato_peel_options.T is not None:
        if not Path(potato_peel_options.T).exists():
            create_directory(directory=Path(potato_peel_options.T))
        bind_dirs.append(Path(potato_peel_options.T))

    # Now run the command and hope foe the best you silly pirate
    run_singularity_command(
        image=potato_container, command=potato_peel_command.command, bind_dirs=bind_dirs
    )

    return potato_peel_command


class NormalisedSources(NamedTuple):
    """Temporary container to hold the normalised source properties that
    would be provided to potato
    """

    source_ras: tuple[float]
    """The RAs in degrees"""
    source_decs: tuple[float]
    """The Decs in degrees"""
    source_fovs: tuple[float]
    """The size of each source to image in degrees"""
    source_names: tuple[str]
    """The name of each source"""


def get_source_props_from_table(table: Table) -> NormalisedSources:
    """Given the astropy table of known sources to peel, normalise their
    inputs for the potato peel CLI

    Args:
        table (Table): Table of sources to peel

    Returns:
        NormalisedSources: Collection of normalised sources properties that will be provided to `hot_potato`
    """

    sources_sky = SkyCoord(table["RA"], table["Dec"], unit=(u.deg, u.deg))

    source_ras = [source_sky.ra.deg for source_sky in sources_sky]  # type: ignore
    source_decs = [source_sky.dec.deg for source_sky in sources_sky]  # type: ignore
    source_apertures = (size for size in (table["Aperture"] * u.arcmin).to(u.deg).value)  # type: ignore
    source_names = [i.replace(" ", "_") for i in table["Name"].value]  # type: ignore

    return NormalisedSources(
        source_ras=tuple(source_ras),
        source_decs=tuple(source_decs),
        source_fovs=tuple(source_apertures),
        source_names=tuple(source_names),
    )


def _print_ms_colnames(ms: MS) -> MS:
    """A dummy function to print colnames in a MS table"""
    ms = MS.cast(ms=ms)

    with table(str(ms.path)) as tab:
        colnames = tab.colnames()

    logger.debug(f"The MS column names are: {colnames=}")

    return ms


def potato_peel(
    ms: MS,
    potato_container: Path,
    update_potato_config_options: dict[str, Any] | None = None,
    update_potato_peel_options: dict[str, Any] | None = None,
    image_options: WSCleanOptions | None = None,
) -> MS:
    """Peel out sources from a measurement set using PotatoPeel. Candidate sources
    from a known list of sources (see Table 3 or RACS-Mid paper) are considered.

    Args:
        ms (MS): The measurement set to peel out known sources
        potato_container (Path): Location of container with potatopeel software installed
        update_potato_config_options (Optional[Dict[str, Any]], optional): A dictionary with values to use to update the default options within the `PotatoConfigOptions`. If None use the defaults. Defaults to None.
        update_potato_peel_options (Optional[Dict[str, Any]], optional): A dictionary with values to use to update the default options within the `PotatoPeelOptions`. If None use the defaults. Defaults to None.
        image_options (Optional[WSCleanOptions], optional): Any imaging options that should be used to determine if sources require peeling (e.g. image size, pixel size)

    Returns:
        MS: Updated measurement set
    """
    potato_container = potato_container.absolute()

    logger.info(f"Will attempt to peel the {ms=}")
    logger.info(f"Using the potato peel container {potato_container}")

    if image_options is None:
        logger.info("No supplied image options, using default WSCleanOptions()")
        image_options = WSCleanOptions()

    logger.info("Colnames before potato")
    _print_ms_colnames(ms=ms)

    peel_tab = find_sources_to_peel(
        ms=ms, image_options=image_options, maximum_offset=6
    )

    if peel_tab is None or len(peel_tab) == 0:
        logger.info("No sources to peel. ")
        return ms

    logger.info(f"Will be peeling {len(peel_tab)} objects: {peel_tab['Name']}")

    update_potato_config_options = (
        update_potato_config_options if update_potato_config_options else {}
    )
    potato_config_options = PotatoConfigOptions(**update_potato_config_options)
    potato_config_command = create_run_potato_config(
        potato_container=potato_container,
        ms_path=ms,
        potato_config_options=potato_config_options,
    )

    ms = prepare_ms_for_potato(ms=ms)

    normalised_source_props = get_source_props_from_table(table=peel_tab)
    potato_peel_arguments = PotatoPeelArguments(
        ms=ms.path,
        ras=normalised_source_props.source_ras,
        decs=normalised_source_props.source_decs,
        peel_fovs=normalised_source_props.source_fovs,
        image_fov=0.01,
        n=normalised_source_props.source_names,
    )

    potato_peel_options = PotatoPeelOptions(
        c=potato_config_command.config_path,
    )
    if update_potato_peel_options:
        potato_peel_options = potato_peel_options.with_options(
            **update_potato_peel_options
        )

    create_run_potato_peel(
        potato_container=potato_container,
        ms=ms,
        potato_peel_arguments=potato_peel_arguments,
        potato_peel_options=potato_peel_options,
    )

    logger.info("Column names after potato")
    _print_ms_colnames(ms=ms)

    return ms.with_options(column="DATA")


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="A simple interface into the potatopeel software"
    )

    subparser = parser.add_subparsers(dest="mode")

    check_parser = subparser.add_parser(
        "check", help="Check whether there are sources to peel from a measurement set"
    )

    check_parser.add_argument(
        "ms", type=Path, help="Path to the measurement set that will be considered"
    )
    check_parser.add_argument(
        "--image-size",
        type=int,
        default=8000,
        help="The number of pixels that make up a square image. Used to determine if a source is within FoV",
    )
    check_parser.add_argument(
        "--pixel-scale",
        type=str,
        default="2.5arcsec",
        help="The size of a pixel in an astropy-understood unit. Used to assess whether a source is within the image FoV. ",
    )

    list_parser = subparser.add_parser(  # noqa
        "list", help="List the known candidate sources available for peeling. "
    )

    peel_parser = subparser.add_parser(
        "peel", help="Attempt to peel sources using potatopeel"
    )
    peel_parser.add_argument(
        "ms", type=Path, help="Path the to the measurement set with a source to peel"
    )

    peel_parser.add_argument(
        "--potato-container",
        type=Path,
        default=Path("./potato.sif"),
        help="Path to the singularity container that represented potatopeel",
    )
    peel_parser.add_argument(
        "--data-column",
        type=str,
        default="DATA",
        help="The column name that contains data with a source that needs to be peeled",
    )
    peel_parser.add_argument(
        "--image-size",
        type=int,
        default=8000,
        help="The number of pixels that make up a square image. Used to determine if a source is within FoV",
    )
    peel_parser.add_argument(
        "--pixel-scale",
        type=str,
        default="2.5arcsec",
        help="The size of a pixel in an astropy-understood unit. Used to assess whether a source is within the image FoV. ",
    )

    return parser


def cli():
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "list":
        tab = load_known_peel_sources()
        logger.info("Known candidate sources to peel")
        logger.info(tab)

    elif args.mode == "check":
        ms = MS(path=args.ms)
        image_options = WSCleanOptions(size=args.image_size, scale=args.pixel_scale)

        tab = find_sources_to_peel(ms=ms, image_options=image_options)
        logger.info("Sources to peel")
        logger.info(tab)

    elif args.mode == "peel":
        ms = MS(path=args.ms, column=args.data_column)
        image_options = WSCleanOptions(size=args.image_size, scale=args.pixel_scale)

        potato_peel(
            ms=ms, potato_container=args.potato_container, image_options=image_options
        )

    else:
        parser.print_help()


if __name__ == "__main__":
    cli()



"""Some tests related to components around measurement sets."""

from __future__ import annotations

from datetime import datetime
from pathlib import Path, PosixPath

import pytest

from flint.exceptions import NamingException
from flint.naming import (
    CASDANameComponents,
    FITSMaskNames,
    ProcessedNameComponents,
    RawNameComponents,
    _rename_linear_to_stokes,
    add_timestamp_to_path,
    casda_ms_format,
    create_fits_mask_names,
    create_image_cube_name,
    create_imaging_name_prefix,
    create_linmos_base_path,
    create_linmos_names,
    create_ms_name,
    create_name_from_common_fields,
    create_path_from_processed_name_components,
    extract_beam_from_name,
    extract_components_from_name,
    get_aocalibrate_output_path,
    get_beam_resolution_str,
    get_fits_cube_from_paths,
    get_potato_output_base_path,
    get_sbid_from_path,
    get_selfcal_ms_name,
    processed_ms_format,
    raw_ms_format,
    rename_linear_to_stokes,
    split_images,
    update_beam_resolution_field_in_path,
)
from flint.options import MS


def test_create_path_from_process_named_components():
    """Make sure we can create a name"""
    components = ProcessedNameComponents(
        sbid="39400", field="RACS_0000-123", beam="33", spw=None, round="3", pol="i"
    )
    assert isinstance(components, ProcessedNameComponents)

    ex = Path("SB39400.RACS_0000-123.beam33.round3.i")
    parent = Path("Jack/Sparrow/Pirate/King")
    out1 = create_path_from_processed_name_components(
        processed_name_components=components
    )
    out2 = create_path_from_processed_name_components(
        processed_name_components=components, parent_path=parent
    )
    assert isinstance(out1, Path)
    assert isinstance(out2, Path)
    assert ex == out1
    assert parent / ex == out2

    components = ProcessedNameComponents(
        sbid="39400", field="RACS_0000-123", beam="33", spw=234, round="3", pol="i"
    )
    assert isinstance(components, ProcessedNameComponents)

    ex = Path("SB39400.RACS_0000-123.beam33.spw234.round3.i")
    parent = Path("Jack/Sparrow/Pirate/King")
    out1 = create_path_from_processed_name_components(
        processed_name_components=components
    )
    out2 = create_path_from_processed_name_components(
        processed_name_components=components, parent_path=parent
    )
    assert isinstance(out1, Path)
    assert isinstance(out2, Path)
    assert ex == out1
    assert parent / ex == out2

    components = ProcessedNameComponents(
        sbid="39400",
        field="RACS_0000-123",
        beam="33",
        spw=234,
        round="3",
        pol="i",
        channel_range=(123, 567),
    )
    assert isinstance(components, ProcessedNameComponents)

    ex = Path("SB39400.RACS_0000-123.beam33.spw234.round3.i.ch0123-0567")
    parent = Path("Jack/Sparrow/Pirate/King")
    out1 = create_path_from_processed_name_components(
        processed_name_components=components
    )
    out2 = create_path_from_processed_name_components(
        processed_name_components=components, parent_path=parent
    )
    assert isinstance(out1, Path)
    assert isinstance(out2, Path)
    assert ex == out1
    assert parent / ex == out2

    components = ProcessedNameComponents(
        sbid="39400",
        field="RACS_0000-123",
        beam="33",
        spw=234,
        round="3",
        pol="i",
        channel_range=(123, 567444),
    )
    assert isinstance(components, ProcessedNameComponents)

    ex = Path("SB39400.RACS_0000-123.beam33.spw234.round3.i.ch0123-567444")
    parent = Path("Jack/Sparrow/Pirate/King")
    out1 = create_path_from_processed_name_components(
        processed_name_components=components
    )
    out2 = create_path_from_processed_name_components(
        processed_name_components=components, parent_path=parent
    )
    assert isinstance(out1, Path)
    assert isinstance(out2, Path)
    assert ex == out1
    assert parent / ex == out2


def test_create_path_from_process_named_components_2():
    """Make sure we can create a name. The one makes sure we can go full circle"""
    parent = Path("Jacccckkkk/Sparrow")
    ex = parent / Path("SB39400.RACS_0000-123.beam33.spw234.round3.i.ch0123-567444")
    pcn = processed_ms_format(in_name=ex)
    out = create_path_from_processed_name_components(
        processed_name_components=pcn, parent_path=parent
    )
    assert ex == out

    parent = Path("Jacccckkkk/Sparrow")
    ex = parent / Path("SB39400.RACS_0000-123.round3.i.ch0123-0444")
    pcn = processed_ms_format(in_name=ex)
    out = create_path_from_processed_name_components(
        processed_name_components=pcn, parent_path=parent
    )
    assert ex == out


def test_create_imaging_name_prefix():
    """Creates the name that will be used for output image
    products"""
    ms = MS.cast(ms=Path("/Jack/Sparrow/SB63789.EMU_1743-51.beam03.round4.ms"))

    name = create_imaging_name_prefix(ms_path=ms.path)
    assert name == "SB63789.EMU_1743-51.beam03.round4"

    for pol in ("I", "i"):
        name = create_imaging_name_prefix(ms_path=ms.path, pol=pol)
        assert name == "SB63789.EMU_1743-51.beam03.round4.i"

        name = create_imaging_name_prefix(
            ms_path=ms.path, pol=pol, channel_range=(100, 108)
        )
        assert name == "SB63789.EMU_1743-51.beam03.round4.i.ch0100-0108"

    name = create_imaging_name_prefix(ms_path=ms.path, channel_range=(100, 108))
    assert name == "SB63789.EMU_1743-51.beam03.round4.ch0100-0108"


def test_get_cube_fits_from_paths():
    """Identify the files that contain the cube field and are fits"""
    files = [
        "SB63789.EMU_1743-51.beam03.round4.i.image.cube.fits",
        "SB63789.EMU_1743-51.beam03.round4.i.image.cube.other.fields.fits",
        "SB63789.EMU_1743-51.beam03.round4.i.MFS.image.optimal.conv.fits",
        "SB63789.EMU_1743-51.beam03.round4.i.MFS.residual.optimal.conv.fits",
        "SB63789.EMU_1743-51.beam03.round4.i.MFS.image.fits",
        "SB63789.EMU_1743-51.beam03.round4.i.MFS.residual.fits",
    ]
    files = [Path(f) for f in files]

    cube_files = get_fits_cube_from_paths(paths=files)

    assert len(cube_files) == 2
    assert cube_files[0] == Path("SB63789.EMU_1743-51.beam03.round4.i.image.cube.fits")
    assert cube_files[1] == Path(
        "SB63789.EMU_1743-51.beam03.round4.i.image.cube.other.fields.fits"
    )


def test_create_image_cube_name():
    """Put together a consistent file cube name"""
    name = create_image_cube_name(
        image_prefix=Path(
            "/jack/sparrow/worst/pirate/flint_fitscube/57222/SB57222.RACS_1141-55.beam10.round3.i"
        ),
        mode="image",
    )
    assert isinstance(name, Path)
    assert name == Path(
        "/jack/sparrow/worst/pirate/flint_fitscube/57222/SB57222.RACS_1141-55.beam10.round3.i.image.cube.fits"
    )

    name = create_image_cube_name(
        image_prefix=Path("./57222/SB57222.RACS_1141-55.beam10.round3.i"),
        mode="residual",
    )
    assert isinstance(name, Path)
    assert name == Path(
        "./57222/SB57222.RACS_1141-55.beam10.round3.i.residual.cube.fits"
    )

    name = create_image_cube_name(
        image_prefix=Path("./57222/SB57222.RACS_1141-55.beam10.round3.i"),
        mode=["residual", "pirate", "imaging"],
    )
    assert isinstance(name, Path)
    assert name == Path(
        "./57222/SB57222.RACS_1141-55.beam10.round3.i.residual.pirate.imaging.cube.fits"
    )
    name = create_image_cube_name(
        image_prefix=Path("./57222/SB57222.RACS_1141-55.beam10.round3.i"),
        mode=["residual", "pirate", "imaging"],
        suffix="jackie",
    )
    assert isinstance(name, Path)
    assert name == Path(
        "./57222/SB57222.RACS_1141-55.beam10.round3.i.residual.pirate.imaging.jackie.cube.fits"
    )
    name = create_image_cube_name(
        image_prefix=Path("./57222/SB57222.RACS_1141-55.beam10.round3.i"),
        mode=["residual", "pirate", "imaging"],
        suffix=["jackie", "boi"],
    )
    assert isinstance(name, Path)
    assert name == Path(
        "./57222/SB57222.RACS_1141-55.beam10.round3.i.residual.pirate.imaging.jackie.boi.cube.fits"
    )


def test_get_beam_resolution_str():
    """Map the known / support modes of beam resolution in file names"""
    assert "raw" == get_beam_resolution_str(mode="raw")
    assert "optimal" == get_beam_resolution_str(mode="optimal")
    assert "fixed" == get_beam_resolution_str(mode="fixed")

    assert "raw!" == get_beam_resolution_str(mode="raw", marker="!")
    assert "optimal?" == get_beam_resolution_str(mode="optimal", marker="?")
    assert "fixed." == get_beam_resolution_str(mode="fixed", marker=".")

    with pytest.raises(ValueError):
        _ = get_beam_resolution_str("Jack")


def test_update_beam_resolution_mode_in_path():
    """Given a path that has a known beam resolution mode in it, update to another"""

    example = Path("SB57516.RACS_0929-81.round4.i.optimal.round4.residual.linmos.fits")
    expected = Path("SB57516.RACS_0929-81.round4.i.fixed.round4.residual.linmos.fits")

    assert expected == update_beam_resolution_field_in_path(
        path=example, original_mode="optimal", updated_mode="fixed"
    )
    assert expected == update_beam_resolution_field_in_path(
        path=example, original_mode="optimal", updated_mode="fixed", marker="."
    )
    with pytest.raises(AssertionError):
        update_beam_resolution_field_in_path(
            path=example, original_mode="fixed", updated_mode="optimal"
        )
        assert expected == update_beam_resolution_field_in_path(
            path=example, original_mode="optimal", updated_mode="fixed", marker="!"
        )
        update_beam_resolution_field_in_path(
            path=Path("JackSparrowCaresNotForBeamResolutions"),
            original_mode="optimal",
            updated_mode="fixed",
        )


def test_casda_ms_format_1934():
    """Checks around the name format form CASDA  considering 1934"""
    exs = [
        "1934.SB40470.beam35.ms",
        Path("1934.SB40470.beam35.ms"),
    ]
    for ex in exs:
        res = casda_ms_format(in_name=ex)
        assert res is not None
        assert res.format == "1934"
        assert res.sbid == 40470
        assert res.beam == "35"


def test_casda_ms_format():
    """Checks around the name format form CASDA"""
    exs = [
        "scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms",
        Path(
            "scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms"
        ),
    ]
    for ex in exs:
        res = casda_ms_format(in_name=ex)
        assert res is not None
        assert isinstance(res, CASDANameComponents)
        assert res.sbid == 40470
        assert res.beam == "35"
        assert res.field == "RACS_1237+00"
        assert res.alias == "RACS_1237+00"
        assert res.format == "scienceData"

    # Confirm None is returned in silly cases
    exs = [
        "scienceData.fdgdfdfg.RACS_1237+00.SBaveraged_cal.leakage.ms",
        "SB12349.RACS_1234+45.ms",
        "SB12349.RACS_1234+45.round2.ms",
    ]
    for ex in exs:
        res = casda_ms_format(in_name=ex)
        assert res is None


def test_self_cal_name_wbeams():
    """Checks around where the self-calibration naming is working, and the
    correct slicing. This test came after the one below and includes the beam
    number as a component to consider"""

    for beam in range(45):
        ms = Path(f"SB12349.RACS_1234+45.beam{beam:02d}.ms")
        e_ms = Path(f"SB12349.RACS_1234+45.beam{beam:02d}.round1.ms")
        out_ms = get_selfcal_ms_name(in_ms_path=ms, round=1)
        assert out_ms == e_ms

    for round in range(1, 5):
        for beam in range(45):
            ms = Path(f"SB12349.RACS_1234+45.beam{beam:02d}.round{round}.ms")
            e_ms = Path(f"SB12349.RACS_1234+45.beam{beam:02d}.round{round + 1}.ms")
            out_ms = get_selfcal_ms_name(in_ms_path=ms, round=round + 1)
            assert out_ms == e_ms


def test_self_cal_name():
    """Checks around where the self-calibration naming is working, and the
    correct slicing"""

    ms = Path("SB12349.RACS_1234+45.ms")
    e_ms = Path("SB12349.RACS_1234+45.round1.ms")
    out_ms = get_selfcal_ms_name(in_ms_path=ms, round=1)
    assert out_ms == e_ms

    ms = Path("SB12349.RACS_1234+45.round1.ms")
    e_ms = Path("SB12349.RACS_1234+45.round2.ms")
    out_ms = get_selfcal_ms_name(in_ms_path=ms, round=2)
    assert out_ms == e_ms

    ms = Path(
        "/some/other/directory/SB12349.RACS_1234+45.verrrrryyylonnnnnnnhhhhh.round1.ms"
    )
    e_ms = Path(
        "/some/other/directory/SB12349.RACS_1234+45.verrrrryyylonnnnnnnhhhhh.round2.ms"
    )
    out_ms = get_selfcal_ms_name(in_ms_path=ms, round=2)
    assert out_ms == e_ms


def test_potato_output_name():
    ms = Path("/some/made/up/path/SB123.Tim.beam33.round2.ms")
    potato = get_potato_output_base_path(ms_path=ms)

    assert isinstance(potato, Path)
    assert potato == Path("/some/made/up/path/SB123.Tim.beam33.potato")

    config = potato.parent / (potato.name + ".config")
    assert config == Path("/some/made/up/path/SB123.Tim.beam33.potato.config")


def test_add_timestamp_to_path():
    """make sure adding a timestamp to a file name works"""
    dd = datetime(2024, 4, 12, 10, 30, 50, 243910)

    example_str = "/test/this/is/filename.txt"
    stamped_path = add_timestamp_to_path(input_path=example_str, timestamp=dd)
    expected = Path("/test/this/is/filename-20240412-103050.txt")

    assert stamped_path == expected

    example_path = Path("/test/this/is/filename.txt")
    stamped_path = add_timestamp_to_path(input_path=example_path, timestamp=dd)

    assert stamped_path == expected

    now_path = add_timestamp_to_path(input_path=example_path)
    assert now_path != expected


def test_create_fits_mask_names_no_signal():
    fits_image = Path("38960/SB38960.RACS_1418-12.noselfcal_linmos.fits")

    result = create_fits_mask_names(fits_image=fits_image, include_signal_path=False)
    assert isinstance(result, FITSMaskNames)
    assert isinstance(result.mask_fits, Path)
    assert result.signal_fits is None
    assert result.mask_fits == Path(
        "38960/SB38960.RACS_1418-12.noselfcal_linmos.mask.fits"
    )


def test_create_fits_mask_names():
    fits_image = Path("38960/SB38960.RACS_1418-12.noselfcal_linmos.fits")

    result = create_fits_mask_names(fits_image=fits_image, include_signal_path=True)
    assert isinstance(result, FITSMaskNames)
    assert isinstance(result.signal_fits, Path)
    assert isinstance(result.mask_fits, Path)
    assert result.signal_fits == Path(
        "38960/SB38960.RACS_1418-12.noselfcal_linmos.signal.fits"
    )
    assert result.mask_fits == Path(
        "38960/SB38960.RACS_1418-12.noselfcal_linmos.mask.fits"
    )


def test_aocalibrate_into_process_ms_format():
    ms = Path("/some/made/up/path/SB123.Tim.beam33.round2.ms")

    outpath = get_aocalibrate_output_path(
        ms_path=ms, include_preflagger=True, include_smoother=True
    )
    named_components = processed_ms_format(in_name=outpath)
    assert isinstance(named_components, ProcessedNameComponents)
    assert named_components.sbid == "123"
    assert named_components.field == "Tim"
    assert named_components.beam == "33"


def test_aocalibrate_naming():
    ms = Path("/some/made/up/path/SB123.Tim.beam33.round2.ms")

    outpath = get_aocalibrate_output_path(
        ms_path=ms, include_preflagger=True, include_smoother=True
    )
    assert outpath == Path(
        "/some/made/up/path/SB123.Tim.beam33.aocalibrate.preflagged.smoothed.bin"
    )

    outpath = get_aocalibrate_output_path(
        ms_path=ms, include_preflagger=True, include_smoother=False
    )
    assert outpath == Path(
        "/some/made/up/path/SB123.Tim.beam33.aocalibrate.preflagged.bin"
    )

    outpath = get_aocalibrate_output_path(
        ms_path=ms, include_preflagger=False, include_smoother=False
    )
    assert outpath == Path("/some/made/up/path/SB123.Tim.beam33.aocalibrate.bin")

    processed_outputs = processed_ms_format(in_name=outpath)
    assert isinstance(processed_outputs, ProcessedNameComponents)

    ms = Path("/some/made/up/path/SB123.Tim.beam33.spw1.round2.ms")

    outpath = get_aocalibrate_output_path(
        ms_path=ms, include_preflagger=True, include_smoother=True
    )
    assert outpath == Path(
        "/some/made/up/path/SB123.Tim.beam33.spw1.aocalibrate.preflagged.smoothed.bin"
    )


def test_get_sbid_from_path():
    example_path = "/scratch3/gal16b/askap_sbids/39400/2022-04-14_100122_0.ms"

    sbid = get_sbid_from_path(path=Path(example_path))
    assert isinstance(sbid, int)
    assert sbid == 39400

    example_path_2 = "/scratch3/gal16b/askap_sbids/39400"
    sbid = get_sbid_from_path(path=Path(example_path_2))
    assert isinstance(sbid, int)
    assert sbid == 39400

    example_path_3 = (
        "/scratch3/gal16b/askap_sbids/39400/2022-04-14_100122_0.averaged.ms"
    )

    sbid = get_sbid_from_path(path=Path(example_path_3))
    assert isinstance(sbid, int)
    assert sbid == 39400


def test_raw_name_components():
    ms = "2022-04-14_100122_0.ms"

    components = raw_ms_format(ms)
    assert isinstance(components, RawNameComponents)
    assert components.beam == "0"
    assert components.date == "2022-04-14"
    assert components.time == "100122"
    assert components.spw is None

    ms = "2022-04-14_100122_0_3.ms"

    components = raw_ms_format(ms)
    assert isinstance(components, RawNameComponents)
    assert components.beam == "0"
    assert components.date == "2022-04-14"
    assert components.time == "100122"
    assert components.spw == "3"


def test_components_all_beams_spws():
    for beam in range(36):
        for spw in range(10):
            ms = f"2022-04-14_100122_{beam}_{spw}.ms"

            components = raw_ms_format(ms)
            assert isinstance(components, RawNameComponents)
            assert components.beam == f"{beam}"
            assert components.date == "2022-04-14"
            assert components.time == "100122"
            assert components.spw == f"{spw}"

    for beam in range(36):
        ms = f"2022-04-14_100122_{beam}.ms"

        components = raw_ms_format(ms)
        assert isinstance(components, RawNameComponents)
        assert components.beam == f"{beam}"
        assert components.date == "2022-04-14"
        assert components.time == "100122"
        assert components.spw is None


def test_create_ms_name():
    example_path = "/scratch3/gal16b/askap_sbids/39400/2022-04-14_100122_0.ms"
    expected = "SB39400.beam00.ms"
    ms_path = create_ms_name(ms_path=Path(example_path))
    assert isinstance(ms_path, str)
    assert ms_path == expected
    assert ms_path.endswith(".ms")

    example_path_2 = "/scratch3/gal16b/askap_sbids/39400/2022-04-14_100122_0.ms"
    expected_2 = "SB39400.RACS_0635-31.beam00.ms"
    ms_path_2 = create_ms_name(ms_path=Path(example_path_2), field="RACS_0635-31")
    assert isinstance(ms_path_2, str)
    assert ms_path_2 == expected_2

    example_path_3 = "/scratch3/gal16b/askap_sbids/39400/2022-04-14_100122_0.ms"
    expected_3 = "SB1234.RACS_0635-31.beam00.ms"
    ms_path_3 = create_ms_name(
        ms_path=Path(example_path_3), sbid=1234, field="RACS_0635-31"
    )
    assert isinstance(ms_path_3, str)
    assert ms_path_3 == expected_3

    example_path_4 = "/scratch3/gal16b/askap_sbids/39400/2022-04-14_100122_0_12.ms"
    expected_4 = "SB1234.RACS_0635-31.beam00.spw12.ms"
    ms_path_4 = create_ms_name(
        ms_path=Path(example_path_4), sbid=1234, field="RACS_0635-31"
    )
    assert isinstance(ms_path_4, str)
    assert ms_path_4 == expected_4

    examples = [
        "scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms",
    ]
    for ex in examples:
        name = create_ms_name(ms_path=Path(ex))
        assert name == "SB40470.RACS_1237+00.beam35.ms"


def test_create_ms_name_no_sbid():
    example_path = "2022-04-14_100122_0.ms"
    expected = "SB0000.beam00.ms"
    ms_path = create_ms_name(ms_path=example_path)
    assert isinstance(ms_path, str)
    assert ms_path == expected

    expected_2 = "SB0000.RACS_0635-31.beam00.ms"
    ms_path_2 = create_ms_name(ms_path=example_path, field="RACS_0635-31")
    assert isinstance(ms_path_2, str)
    assert ms_path_2 == expected_2

    example_path_3 = "2022-04-14_100122_0_12.ms"
    expected_3 = "SB0000.RACS_0635-31.beam00.spw12.ms"
    ms_path_3 = create_ms_name(ms_path=example_path_3, field="RACS_0635-31")
    assert isinstance(ms_path_3, str)
    assert ms_path_3 == expected_3


def test_formatted_name_components_withpol():
    """Tests around the pol field in a file name"""
    ex = "SB39400.RACS_0635-31.beam33-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round is None
    assert components.pol is None

    ex = "SB39400.RACS_0635-31.beam33.i-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round is None
    assert components.pol == "i"

    ex = "SB39400.RACS_0635-31.beam33.round2.i-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round == "2"
    assert components.pol == "i"

    ex = "SB39400.RACS_0635-31.beam33.round2.iq-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round == "2"
    assert components.pol == "iq"

    ex = "SB39400.RACS_0635-31.beam33.round2-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round == "2"
    assert components.pol is None


def test_formatted_name_components():
    ex = "SB39400.RACS_0635-31.beam33-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round is None

    ex = "SB39400.RACS_0635-31.beam33.spw123-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw == "123"
    assert components.round is None

    ex_path = Path(
        "/this/is/and/example/path/SB39400.RACS_0635-31.beam33.spw123-MFS-image.conv.fits"
    )

    components = processed_ms_format(in_name=ex_path)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw == "123"
    assert components.round is None


def test_formatted_name_components_wchannelrange():
    ex = "SB39400.RACS_0635-31.beam33.round1.i.ch0100-1009.MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round == "1"
    assert components.pol == "i"
    assert components.channel_range == (100, 1009)

    ex = "SB39400.RACS_0635-31.beam33.round1.ch0100-1009.MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round == "1"
    assert components.pol is None
    assert components.channel_range == (100, 1009)


def test_formatted_name_components_wround():
    ex = "SB39400.RACS_0635-31.beam33.round1-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw is None
    assert components.round == "1"

    ex = "SB39400.RACS_0635-31.beam33.spw123.round12-MFS-image.conv.fits"

    components = processed_ms_format(in_name=ex)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw == "123"
    assert components.round == "12"

    ex_path = Path(
        "/this/is/and/example/path/SB39400.RACS_0635-31.beam33.spw123.round123-MFS-image.conv.fits"
    )

    components = processed_ms_format(in_name=ex_path)
    assert isinstance(components, ProcessedNameComponents)
    assert components.sbid == "39400"
    assert components.field == "RACS_0635-31"
    assert components.beam == "33"
    assert components.spw == "123"
    assert components.round == "123"


def test_get_correct_name_format():
    examples = [
        "SB39400.RACS_0635-31.beam33-MFS-image.conv.fits",
        "SB39400.RACS_0635-31.beam33.ms",
        Path(
            "/this/is/and/example/path/SB39400.RACS_0635-31.beam33.spw123-MFS-image.conv.fits"
        ),
    ]

    for ex in examples:
        res = extract_components_from_name(name=ex)
        assert isinstance(res, ProcessedNameComponents)

    examples = ["2022-04-14_100122_0.averaged.ms"]

    for ex in examples:
        res = extract_components_from_name(name=ex)
        assert isinstance(res, RawNameComponents)

    examples = [
        "scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms",
        Path(
            "scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms"
        ),
    ]
    for ex in examples:
        res = extract_components_from_name(name=ex)
        assert isinstance(res, CASDANameComponents)


def test_get_beam_from_name():
    assert extract_beam_from_name(name="2022-04-14_100122_0.averaged.ms") == 0
    assert (
        extract_beam_from_name(name="SB39400.RACS_0635-31.beam33-MFS-image.conv.fits")
        == 33
    )


def get_lots_of_names() -> list[Path]:
    examples = [
        "59058/SB59058.RACS_1626-84.ch0285-0286.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0285-0286.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0070-0071.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0142-0143.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0214-0215.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0286-0287.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0071-0072.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0143-0144.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0215-0216.linmos.fits",
        "59058/SB59058.RACS_1626-84.ch0287-0288.linmos.fits",
    ]

    return list(map(Path, examples))


def test_create_name_from_common_fields():
    """See if we can identify the common bits of names as recognised in the process name format"""
    examples = get_lots_of_names()

    common_names = create_name_from_common_fields(in_paths=examples)
    expected_common_name = Path("59058/SB59058.RACS_1626-84")

    assert common_names == expected_common_name

    for additional_suffix in (".linmos.fits", "linmos.fits"):
        common_names = create_name_from_common_fields(
            in_paths=examples, additional_suffixes=additional_suffix
        )
        expected_common_name = Path("59058/SB59058.RACS_1626-84.linmos.fits")

        assert common_names == expected_common_name

    examples.append("This/will/raise/a/valuerror")

    with pytest.raises(ValueError):
        create_name_from_common_fields(in_paths=examples)


def get_lots_of_names_2() -> list[Path]:
    examples = [
        "59058/SB59058.RACS_1626-84.round4.i.ch0285-0286.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0285-0286.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0070-0071.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0142-0143.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0214-0215.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0286-0287.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0071-0072.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0143-0144.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0215-0216.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0287-0288.linmos.fits",
    ]

    return list(map(Path, examples))


def test_create_name_from_common_fields_2():
    """See if we can identify the common bits of names as recognised in the process name format.
    This picks up some missing formats that this sea dog initially overlookede"""
    examples = get_lots_of_names_2()

    common_names = create_name_from_common_fields(in_paths=examples)
    expected_common_name = Path("59058/SB59058.RACS_1626-84.round4.i")

    assert common_names == expected_common_name

    for additional_suffix in (".linmos.fits", "linmos.fits"):
        common_names = create_name_from_common_fields(
            in_paths=examples, additional_suffixes=additional_suffix
        )
        expected_common_name = Path("59058/SB59058.RACS_1626-84.round4.i.linmos.fits")

        assert common_names == expected_common_name

    examples.append("This/will/raise/a/valuerror")

    with pytest.raises(ValueError):
        create_name_from_common_fields(in_paths=examples)


def get_lots_of_names_3():
    files = [
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam00.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam01.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam02.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam03.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam04.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam05.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam06.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam07.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
        PosixPath(
            "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.beam08.round4.i.ch0238-0239.image.optimal.conv.fits"
        ),
    ]

    return files


def test_create_name_from_common_fields_3():
    """Test a strange combination of errors around the common field name"""

    files = get_lots_of_names_3()

    common_name = create_name_from_common_fields(in_paths=files)
    assert common_name == PosixPath(
        "/scratch3/gal16b/spectral_flow/57516/SB57516.RACS_0929-81.round4.i.ch0238-0239"
    )


def test_create_linmos_parset_base_path():
    """The yandasoft linmos task writes out a configuration file.
    This function tests the generation of the path"""
    examples = get_lots_of_names_2()

    expected = Path("59058/SB59058.RACS_1626-84.round4.i").absolute()
    assert expected == create_linmos_base_path(input_images=examples)

    expected = Path("59058/SB59058.RACS_1626-84.round4.i.jack.sparrow").absolute()
    assert expected == create_linmos_base_path(
        input_images=examples, additional_suffixes="jack.sparrow"
    )
    new_paths = [Path("/Here/Be/Pirates") / p for p in examples]
    expected = Path("/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i").absolute()
    assert expected == create_linmos_base_path(input_images=new_paths)

    expected = Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.jack.sparrow"
    ).absolute()
    assert expected == create_linmos_base_path(
        input_images=new_paths, additional_suffixes="jack.sparrow"
    )


def test_create_linmos_names():
    """Some logic is around creating names to give to linmos to create the output images, weights and parset"""

    linmos_names = create_linmos_names(
        name_prefix="/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i"
    )
    assert linmos_names.image_fits == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.linmos.fits"
    )
    assert linmos_names.weight_fits == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.weight.fits"
    )
    assert linmos_names.parset_output_path == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i_parset.txt"
    )

    linmos_names = create_linmos_names(
        name_prefix=Path("/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i")
    )
    assert linmos_names.image_fits == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.linmos.fits"
    )
    assert linmos_names.weight_fits == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.weight.fits"
    )
    assert linmos_names.parset_output_path == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i_parset.txt"
    )

    linmos_names = create_linmos_names(
        name_prefix="/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i",
        parset_output_path=Path("Jack/My/Boi.txt"),
    )
    assert linmos_names.image_fits == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.linmos.fits"
    )
    assert linmos_names.weight_fits == Path(
        "/Here/Be/Pirates/59058/SB59058.RACS_1626-84.round4.i.weight.fits"
    )
    assert linmos_names.parset_output_path == Path("Jack/My/Boi.txt")


def test_rename_linear_to_stokes():
    linear_names = (
        "jack.qu.sparrow",
        "qu.sparrow.jack",
        ".qu.sparrow.jack",
        "sparrow.jack.qu",
    )
    stokes_q_names = (
        "jack.q.sparrow",
        "qu.sparrow.jack",
        ".q.sparrow.jack",
        "sparrow.jack.q",
    )
    stokes_u_names = (
        "jack.u.sparrow",
        "qu.sparrow.jack",
        ".u.sparrow.jack",
        "sparrow.jack.u",
    )

    for stokes, stokes_names in zip(("q", "u"), (stokes_q_names, stokes_u_names)):
        for linear_name, check_name in zip(linear_names, stokes_names):
            stokes_name = _rename_linear_to_stokes(linear_name, stokes)
            assert stokes_name == check_name

    with pytest.raises(NameError):
        _rename_linear_to_stokes("jack.sparrow", "i")
        _rename_linear_to_stokes("jack.sparrow", "v")
        _rename_linear_to_stokes("jack.sparrow", "pearl")

    stokes_path = rename_linear_to_stokes(
        linear_name=Path("jack.qu.sparrow"), stokes="q"
    )
    assert isinstance(stokes_path, Path)
    stokes_path = rename_linear_to_stokes(linear_name="jack.qu.sparrow", stokes="q")
    assert isinstance(stokes_path, str)


def test_split_images():
    with pytest.raises(ValueError):
        split_images(images=[Path("jack.sparrow.i.fits")], by="pol")

    pols = "iquv"
    images = [Path(f"SB1234.FieldName.beam00.round4.{p}.fits") for p in pols]

    split_dict = split_images(images=images, by="pol")

    assert len(split_dict) == 4
    assert split_dict.keys() == set(pols)
    for key, val in split_dict.items():
        assert len(val) == 1
        assert val[0] == Path(f"SB1234.FieldName.beam00.round4.{key}.fits")

    images = [
        Path("SB1234.FieldName.beam00.round4.i.fits"),
        Path("SB1234.FieldName.beam00.round4.i.MFS.fits"),
        Path("SB1234.FieldName.beam00.round4.v.fits"),
    ]

    split_dict = split_images(images=images, by="pol")
    assert len(split_dict) == 2
    assert split_dict.keys() == {"i", "v"}
    assert len(split_dict["i"]) == 2
    assert len(split_dict["v"]) == 1

    sbids = [
        "1234",
        "5678",
        "9012",
    ]

    images = [Path(f"SB{sbid}.FieldName.beam00.round4.i.fits") for sbid in sbids]
    split_dict = split_images(images=images, by="sbid")
    assert len(split_dict) == 3
    assert list(split_dict.keys()) == list(sbids)
    for key, val in split_dict.items():
        assert len(val) == 1
        assert val[0] == Path(f"SB{key}.FieldName.beam00.round4.i.fits")

    with pytest.raises(NamingException):
        split_images(images=images, by="jack")



"""Basic tests around addmodel"""

from __future__ import annotations

from pathlib import Path

import pytest
from pydantic import ValidationError

from flint.predict.addmodel import (
    AddModelOptions,
    add_model_options_to_command,
    get_parser,
)


def test_get_parser():
    """No silly business around creating the parser (including imports)"""

    _ = get_parser()


def test_create_add_model_instance():
    """Creating the addmodeloptions instance"""

    add_model_options = AddModelOptions(
        model_path=Path("Jack_model.txt"),
        ms_path=Path("Sparrow.ms"),
        mode="a",
        datacolumn="DATA",
    )
    assert isinstance(add_model_options, AddModelOptions)

    with pytest.raises(ValidationError):
        AddModelOptions(
            model_path=Path("Jack_model.txt"),
            ms_path=Path("Sparrow.ms"),
            mode="NoExistsAnRaisesError",
            datacolumn="DATA",
        )


def test_generate_add_model_command():
    """Ensure we can actually generate the expected addmodel cli command"""

    add_model_options = AddModelOptions(
        model_path=Path("/jack/sparrow/be/here/SB-sources.txt"),
        ms_path=Path("/jack/sparrow/be/here/SB.ms"),
        mode="c",
        datacolumn="MODEL_DATA",
    )

    add_model_command = add_model_options_to_command(
        add_model_options=add_model_options
    )

    expected_command = "addmodel -datacolumn MODEL_DATA -m c /jack/sparrow/be/here/SB-sources.txt /jack/sparrow/be/here/SB.ms"
    assert add_model_command == expected_command


def test_addmodel_to_command_string():
    """make sure we generate the correct command string"""

    add_model_options = AddModelOptions(
        model_path=Path("Jack_model.txt"),
        ms_path=Path("Sparrow.ms"),
        mode="a",
        datacolumn="DATA",
    )

    command = add_model_options_to_command(add_model_options=add_model_options)
    expected = "addmodel -datacolumn DATA -m a Jack_model.txt Sparrow.ms"

    assert command == expected



"""Prefect tasks around model visibility prediction"""

from __future__ import annotations

from prefect import task

from flint.logging import logger
from flint.ms import MS
from flint.options import AddModelSubtractFieldOptions
from flint.predict.crystalball import CrystalBallOptions


@task
def task_crystalball_to_ms(ms: MS, crystalball_options: CrystalBallOptions) -> MS:
    """Predict model visibilities into a measurement set using a previously constructed
    blackboard sky model. See ``wsclean -save-source-list`. This used the ``crystalball``
    python package, which under the hood taps into the same dask task runner running
    this flow.

    Visibilities are predicted into the MS's ``MODEL_DATA`` column.

    Args:
        ms (MS): The measurement set where model visibilities will be predicted into.
        crystalball_options (CrystalBallOptions): Options around the crystal ball operation

    Returns:
        MS: An updated MS with the model column set
    """
    from prefect_dask import get_dask_client

    from flint.predict.crystalball import crystalball_predict
    from flint.prefect.helpers import enable_loguru_support

    # crystalball uses loguru. We want to try to attach a handler
    enable_loguru_support()

    with get_dask_client() as client:
        logger.info("Obtained the Client supporting the DaskTaskRunner.")
        return crystalball_predict(
            ms=ms,
            crystalball_options=crystalball_options,
            dask_client=client,
            output_column="MODEL_DATA",
        )


@task
def task_addmodel_to_ms(
    ms: MS,
    addmodel_subtract_options: AddModelSubtractFieldOptions,
) -> MS:
    from flint.imager.wsclean import get_wsclean_output_source_list_path
    from flint.predict.addmodel import AddModelOptions, add_model

    logger.info(f"Searching for wsclean source list for {ms.path}")
    for idx, pol in enumerate(addmodel_subtract_options.wsclean_pol_mode):
        wsclean_source_list_path = get_wsclean_output_source_list_path(
            name_path=ms.path, pol=pol
        )
        assert wsclean_source_list_path.exists(), (
            f"{wsclean_source_list_path=} was requested, but does not exist"
        )

        # This should attempt to add model of different polarisations together.
        # But to this point it is a future proof and is not tested.
        addmodel_options = AddModelOptions(
            model_path=wsclean_source_list_path,
            ms_path=ms.path,
            mode="c" if idx == 0 else "a",
            datacolumn="MODEL_DATA",
        )
        assert addmodel_subtract_options.calibrate_container is not None, (
            f"{addmodel_subtract_options.calibrate_container=}, which should not happen"
        )
        add_model(
            add_model_options=addmodel_options,
            container=addmodel_subtract_options.calibrate_container,
            remove_datacolumn=idx == 0,
        )

    return ms.with_options(model_column="MODEL_DATA")



"""Common prefect related utilities that can be used between flows."""

from __future__ import annotations

import base64
from pathlib import Path
from typing import Any, Callable, Iterable, ParamSpec, TypeVar
from uuid import UUID

from prefect import task
from prefect.artifacts import create_markdown_artifact

from flint.archive import copy_sbid_files_archive, create_sbid_tar_archive
from flint.logging import logger
from flint.naming import (
    add_timestamp_to_path,
    get_fits_cube_from_paths,
    get_sbid_from_path,
    rename_linear_to_stokes,
)
from flint.options import ArchiveOptions
from flint.summary import (
    create_beam_summary,
    create_field_summary,
    update_field_summary,
)

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")
C = TypeVar("C", bound=Callable)

SUPPORTED_IMAGE_TYPES = ("png",)


@task
def task_create_object(object: C, **kwargs) -> C:
    return object(**kwargs)


@task
def task_getattr(
    item: object,
    attribute: str,
    /,
) -> Any:
    """Retrieve an attribute from an input instance of a class or structure.

    Args:
        item (Any): The item that has the input class or structure
        attribute (str): The attribute to extract

    Returns:
        Any: Value of the requested attribute
    """
    logger.debug(f"Pulling {attribute=}")
    return getattr(item, attribute)


@task
def task_sorted(
    iterable: Iterable[T],
    /,
    *,
    key: Any = None,
    reverse: bool = False,
) -> list[T]:
    return sorted(iterable, key=key, reverse=reverse)


@task
def task_zip_list_of_list(list_of_list: list[list[T]]) -> list[tuple[T, ...]]:
    return list(zip(*list_of_list))


def upload_image_as_artifact(
    image_path: Path, description: str | None = None
) -> UUID | None:
    """Create and submit a markdown artifact tracked by prefect for an
    input image. Currently supporting png formatted images.

    The input image is converted to a base64 encoding, and embedded directly
    within the markdown string. Therefore, be mindful of the image size as this
    is tracked in the postgres database.

    Args:
        image_path (Path): Path to the image to upload
        description (Optional[str], optional): A description passed to the markdown artifact. Defaults to None.

    Returns:
        UUID | None: Generated UUID of the registered artifact. If the upload fails ``None`` is returned.
    """
    from prefect.exceptions import PrefectHTTPStatusError

    image_type = image_path.suffix.replace(".", "")
    assert image_path.exists(), f"{image_path} does not exist"
    assert image_type in SUPPORTED_IMAGE_TYPES, (
        f"{image_path} has type {image_type}, and is not supported. Supported types are {SUPPORTED_IMAGE_TYPES}"
    )

    with open(image_path, "rb") as open_image:
        logger.info(f"Encoding {image_path} in base64")
        image_base64 = base64.b64encode(open_image.read()).decode()

    logger.info("Creating markdown tag")
    markdown = f"![{image_path.stem}](data:image/{image_type};base64,{image_base64})"

    try:
        image_uuid: UUID | None = None
        logger.info("Registering artifact")
        image_uuid = create_markdown_artifact(
            markdown=markdown, description=description
        )
    except PrefectHTTPStatusError as e:
        logger.warning("Failed to register artefact. ")
        logger.warning(f"{e=}")

    return image_uuid


task_update_field_summary = task(update_field_summary)
task_create_field_summary = task(create_field_summary)
task_create_beam_summary = task(create_beam_summary)
task_get_fits_cube_from_paths = task(get_fits_cube_from_paths)
task_rename_linear_to_stokes = task(rename_linear_to_stokes)


@task
def task_archive_sbid(
    science_folder_path: Path,
    archive_path: Path | None = None,
    copy_path: Path | None = None,
    max_round: int | None = None,
    update_archive_options: dict[str, Any] | None = None,
) -> Path:
    """Create a tarball of files, or copy files, from a processing folder.

    Args:
        science_folder_path (Path): Path that contains the imaged produced
        archive_path (Optional[Path], optional): Location to create and store the tar ball at. If None no tarball is created. Defaults to None.
        copy_path (Optional[Path], optional): Location to copy selected files into. If None no files are copied. Defaults to None.
        max_round (Optional[int], optional): The last self-calibration round performed. If provided some files form this round are copied (assuming wsclean imaging). If None, the default file patterns in ArchiveOptions are used. Defaults to None.
        update_archive_options (Optional[Dict[str, Any]], optional): Additional options to provide to ArchiveOptions. Defaults to None.

    Returns:
        Path: The science folder files were copied from
    """

    sbid = get_sbid_from_path(path=science_folder_path)

    archive_options = ArchiveOptions()

    if update_archive_options:
        logger.info(f"Updating archive options with {update_archive_options=}")
        archive_options = archive_options.with_options(**update_archive_options)

    # TODO: What should this be? Just general new regexs passed through,
    # or is this fine?
    if max_round:
        updated_file_patterns = (
            *tuple(archive_options.tar_file_re_patterns),
            f".*beam[0-9]+\\.round{max_round}-.*-image\\.fits",
            f".*beam[0-9]+\\.round{max_round}\\.ms\\.(zip|tar)",
        )
        archive_options = archive_options.with_options(
            tar_file_re_patterns=updated_file_patterns
        )

    if archive_path:
        tar_file_name = add_timestamp_to_path(Path(archive_path) / f"SB{sbid}.tar")
        create_sbid_tar_archive(
            tar_out_path=tar_file_name,
            base_path=science_folder_path,
            archive_options=archive_options,
        )

    if copy_path:
        copy_sbid_files_archive(
            copy_out_path=copy_path,
            base_path=science_folder_path,
            archive_options=archive_options,
        )

    return science_folder_path


@task
def task_update_with_options(input_object: T, **kwargs) -> T:
    """Updated any object via its `.with_options()` interface.

    All key-word arguments other than `input_object` are passed through
    to that `input_object`s `.with_options()` method.

    Args:
        input_object (T): The object that has an `.with_options` method that will be updated

    Returns:
        T: The updated object
    """
    assert "with_options" in dir(input_object), (
        f"{type(input_object)=} does not have a with_options method"
    )
    updated_object = input_object.with_options(**kwargs)  # type: ignore

    return updated_object


@task
def task_get_attributes(item: Any, attribute: str) -> Any:
    """Retrieve an attribute from an input instance of a class or structure.

    This is intended to be used when dealing with a prefect future object that
    has yet to be evaluated or is otherwise not immediately accessible.

    Args:
        item (Any): The item that has the input class or structure
        attribute (str): The attribute to extract

    Returns:
        Any: Value of the requested attribute
    """
    logger.debug(f"Pulling {attribute=}")
    return item.__dict__[attribute]


@task
def task_flatten(to_flatten: list[list[T]]) -> list[T]:
    """Will flatten a list of lists into a single list. This
    is useful for when a task-descorated function returns a list.


    Args:
        to_flatten (List[List[T]]): Input list of lists to flatten

    Returns:
        List[T]: Flattened form of input
    """
    logger.debug(f"Received {len(to_flatten)} to flatten.")

    flatten_list = [item for sublist in to_flatten for item in sublist]

    logger.debug(f"Flattened list {len(flatten_list)}")

    return flatten_list



"""Basic utilities to load operational parameters from a yaml-based
configuration file. The idea being that a configuration file would
be used to specify the options for imaging and self-calibration
throughout the pipeline.
"""

from __future__ import annotations

import shutil
from argparse import ArgumentParser
from pathlib import Path
from typing import Any

import yaml

from flint.imager.wsclean import WSCleanOptions
from flint.logging import logger
from flint.masking import MaskingOptions
from flint.naming import add_timestamp_to_path
from flint.options import ArchiveOptions
from flint.selfcal.casa import GainCalOptions
from flint.source_finding.aegean import AegeanOptions, BANEOptions

# TODO: It feels like that the standard of this strategy file should
# be updated. In its current form the "initial" section is required, with
# with the subsequent selfcal section being used for continuum imaging.
# Should this be intended to be used for other pipelines it would be
# better to have an "imager" operation mode, and put these two things
# into there

# Known headers must **always** be present in the strategy file
KNOWN_HEADERS = ("defaults", "version")
# Known options are optional, but if present must be in the correct format
KNOWN_OPERATIONS = ("selfcal", "stokesv", "subtractcube", "polarisation")
FORMAT_VERSION = 0.2
MODE_OPTIONS_MAPPING = {
    "wsclean": WSCleanOptions,
    "gaincal": GainCalOptions,
    "masking": MaskingOptions,
    "archive": ArchiveOptions,
    "bane": BANEOptions,
    "aegean": AegeanOptions,
}
POLARISATION_MAPPING = {
    "total": "i",
    "linear": "qu",
    "circular": "v",
}


def _create_mode_mapping_defaults() -> dict[str, Any]:
    """Create the default key-values for each of the registered Option classes

    Returns:
        Dict[str, Any]: Name of mode and the supported keys and default values for each
    """
    return {k: i()._asdict() for k, i in MODE_OPTIONS_MAPPING.items()}


# A simple representation to pass around. Will help the type
# analysis and future pirates be clear with their mutinous
# intentions
class Strategy(dict):
    """Base representation for handling a loaded flint
    strategy"""

    pass


def copy_and_timestamp_strategy_file(output_dir: Path, input_yaml: Path) -> Path:
    """Timestamp and copy the input strategy file to an
    output directory

    Args:
        output_dir (Path): Output directory the file will be copied to
        input_yaml (Path): The file to copy

    Returns:
        Path: Copied and timestamped file path
    """
    stamped_imaging_strategy = (
        output_dir / add_timestamp_to_path(input_path=input_yaml).name
    )
    logger.info(f"Copying {input_yaml.absolute()} to {stamped_imaging_strategy}")
    shutil.copyfile(input_yaml.absolute(), stamped_imaging_strategy)

    return Path(stamped_imaging_strategy)


def load_and_copy_strategy(
    output_split_science_path: Path, imaging_strategy: Path | None = None
) -> Strategy | None:
    """Load a strategy file and copy a timestamped version into the output directory
    that would contain the science processing.

    Args:
        output_split_science_path (Path): Where the strategy file should be copied to (where the data would be processed)
        imaging_strategy (Optional[Path], optional): Location of the strategy file. Defaults to None.

    Returns:
        Union[Strategy, None]: The loadded strategy file if provided, `None` otherwise
    """
    return (
        load_strategy_yaml(
            input_yaml=copy_and_timestamp_strategy_file(
                output_dir=output_split_science_path,
                input_yaml=imaging_strategy,
            ),
            verify=True,
        )
        if imaging_strategy
        else None
    )


def get_selfcal_options_from_yaml(input_yaml: Path | None = None) -> dict:
    """Stub to represent interaction with a configurationf ile

    If a path is supplied, an error is raised.

    Args:
        input_yaml (Optional[Path], optional): Path to the configuration file. . Defaults to Optional[Path]=None.

    Returns:
        Dict: Mapping where the key is the self-calibration round, and values are key-value of updated gaincal options
    """

    assert input_yaml is None, (
        "Configuring via a yaml configuration file is not yet support. "
    )

    return {
        1: {"solint": "60s", "uvrange": ">235m", "nspw": 1},
        2: {"solint": "30s", "calmode": "p", "uvrange": ">235m", "nspw": 4},
        3: {"solint": "60s", "calmode": "ap", "uvrange": ">235m", "nspw": 4},
        4: {"solint": "30s", "calmode": "ap", "uvrange": ">235m", "nspw": 4},
    }


def get_image_options_from_yaml(
    input_yaml: Path | None = None, self_cal_rounds: bool = False
) -> dict:
    """Stub to interact with configuration file.

    If a `input_yaml` file is provided an error is raised

    Args:
        input_yaml (Optional[Path], optional): Should be None. Defaults to None.
        self_cal_rounds (bool, optional): Whether options for first imaging is being provided, or options to supply for each self-cal round. Defaults to False.

    Returns:
        Dict: _description_
    """

    assert input_yaml is None, (
        "Configuring via a yaml configuration file is not yet supported. "
    )

    MULTISCALE_SCALES = (0, 15, 30, 40, 50, 60, 70, 120)
    IMAGE_SIZE = 7144

    if not self_cal_rounds:
        return {
            "size": IMAGE_SIZE,
            "minuvw_m": 235,
            "weight": "briggs -1.5",
            "scale": "2.5arcsec",
            "nmiter": 10,
            "deconvolution_channels": 4,
            "fit_spectral_pol": 3,
            "auto_mask": 10,
            "multiscale": True,
            "local_rms_window": 55,
            "multiscale_scales": MULTISCALE_SCALES,
        }
    else:
        return {
            1: {
                "size": IMAGE_SIZE,
                "weight": "briggs -1.5",
                "scale": "2.5arcsec",
                "nmiter": 20,
                "minuvw_m": 235,
                "channels_out": 18,
                "deconvolution_channels": 6,
                "fit_spectral_pol": 3,
                "auto_mask": 8.0,
                "local_rms_window": 55,
                "multiscale_scales": MULTISCALE_SCALES,
            },
            2: {
                "size": IMAGE_SIZE,
                "weight": "briggs -1.5",
                "scale": "2.5arcsec",
                "multiscale": True,
                "minuvw_m": 235,
                "nmiter": 20,
                "channels_out": 18,
                "deconvolution_channels": 6,
                "fit_spectral_pol": 3,
                "auto_mask": 7.0,
                "local_rms_window": 55,
                "multiscale_scales": MULTISCALE_SCALES,
            },
            3: {
                "size": IMAGE_SIZE,
                "weight": "briggs -1.5",
                "scale": "2.5arcsec",
                "multiscale": True,
                "minuvw_m": 235,
                "nmiter": 20,
                "channels_out": 18,
                "deconvolution_channels": 6,
                "fit_spectral_pol": 3,
                "auto_mask": 6.0,
                "local_rms_window": 55,
                "multiscale_scales": MULTISCALE_SCALES,
            },
            4: {
                "size": IMAGE_SIZE,
                "weight": "briggs -1.5",
                "scale": "2.5arcsec",
                "multiscale": True,
                "minuvw_m": 235,
                "nmiter": 20,
                "channels_out": 16,
                "deconvolution_channels": 4,
                "fit_spectral_pol": 3,
                "auto_mask": 8,
                "local_rms_window": 55,
                "multiscale_scales": MULTISCALE_SCALES,
            },
            5: {
                "size": IMAGE_SIZE,
                "weight": "briggs -1.5",
                "scale": "2.5arcsec",
                "multiscale": True,
                "minuvw_m": 235,
                "nmiter": 20,
                "channels_out": 4,
                "fit_spectral_pol": 3,
                "auto_mask": 7.0,
                "local_rms_window": 55,
                "multiscale_scales": MULTISCALE_SCALES,
            },
        }


def get_options_from_strategy(
    strategy: Strategy | None | Path,
    operation: str,
    mode: str = "wsclean",
    round_info: int | None = None,
    max_round_override: bool = True,
    polarisation: str | None = None,
) -> dict[Any, Any]:
    f"""Extract a set of options from a strategy file to use in a pipeline
    run. If the mode exists in the default section, these are used as a base.

    If the mode exists and a round is specified, the options listed in the
    round are used to update the defaults.

    The default `operation` of `None` implies options related to imaging are
    requested. if a `operation` is provided then options for that mode are
    retrieved. These are ones that do not vary as across rounds of self-calibration.
    An `operation` can have a `mode`, such as `stokesv` requiring a `wsclean` mode. Acceptable
    `operation` values are {KNOWN_OPERATIONS}

    Args:
        strategy (Union[Strategy,None,Path]): A loaded instance of a strategy file. If `None` is provided then an empty dictionary is returned. If `Path` attempt to load the strategy file.
        mode (str, optional): Which set of options to load. Typical values are `wsclean`, `gaincal` and `masking`. Defaults to "wsclean".
        round_info (Union[str, int], optional): Which round to load options for. May be `initial` or an `int` (which indicated a self-calibration round). Defaults to "initial".
        max_round_override (bool, optional): Check whether an integer round number is recorded. If it is higher than the largest self-cal round specified, set it to the last self-cal round. If False this is not performed. Defaults to True.
        operation (Optional[str], optional): Get options related to a specific operation. Defaults to None.

    Raises:
        ValueError: An unrecongised value for `round`.
        AssertError: An unrecongised value for `round`.

    Returns:
        Dict[Any, Any]: Options specific to the requested set
    """

    if strategy is None:
        return {}
    elif isinstance(strategy, Path):
        strategy = load_strategy_yaml(input_yaml=strategy)

    # Some sanity checks
    assert isinstance(strategy, (Strategy, dict)), (
        f"Unknown input strategy type {type(strategy)}"
    )
    assert round_info is None or isinstance(round_info, int), (
        f"{round_info=} not a known value or type. "
    )
    if operation not in KNOWN_OPERATIONS:
        raise ValueError(
            f"{operation=} is not recognised. Known operations are {KNOWN_OPERATIONS}"
        )

    # Override the round if requested
    if (
        isinstance(round_info, int)
        and max_round_override
        and "selfcal" in strategy.keys()
    ):
        round_info = min(round_info, max(strategy["selfcal"].keys()))

    # step one, get the defaults
    options = dict(**strategy["defaults"][mode]) if mode in strategy["defaults"] else {}
    logger.debug(f"Defaults for {mode=}, {options=}")

    # A default empty dict
    update_options = {}

    assert operation in strategy, f"{operation=} not in {strategy.keys()}"

    operation_scope = strategy.get(operation)
    if round_info is not None:
        if not isinstance(operation_scope, dict):
            raise ValueError(
                f"{operation_scope=} is not a dictionary. Cannot extract {round_info=}"
            )
        operation_scope = operation_scope[round_info]

    # Override the polarisation if requested
    if polarisation is not None:
        if not isinstance(operation_scope, dict):
            raise ValueError(
                f"{operation_scope=} is not a dictionary. Cannot extract {polarisation=}"
            )
        operation_scope = operation_scope[polarisation]
        # Update the wsclean options with the polarisation mapping
        if mode == "wsclean":
            operation_scope.setdefault(mode, {}).update(
                {"pol": POLARISATION_MAPPING[polarisation]}
            )

    if not isinstance(operation_scope, dict):
        raise ValueError(
            f"{operation_scope=} is not a dictionary. Cannot extract {round_info=}"
        )

    if mode in operation_scope:
        update_options = dict(**operation_scope[mode])

    if update_options:
        logger.debug(f"Updating options with {update_options=}")
        options.update(update_options)

    return options


def verify_configuration(input_strategy: Strategy, raise_on_error: bool = True) -> bool:
    """Perform basic checks on the configuration file

    Args:
        input_strategy (Strategy): The loaded configuration file structure
        raise_on_error (bool, optional): Whether to raise an error should an issue in thew config file be found. Defaults to True.

    Raises:
        ValueError: Whether structure is valid

    Returns:
        bool: Config file is not valid. Raised only if `raise_on_error` is `True`
    """

    errors: list[str] = []

    for known_header in KNOWN_HEADERS:
        if known_header not in input_strategy.keys():
            errors.append(
                f"Required section header {known_header} missing from input configuration."
            )

    if "version" in input_strategy.keys():
        if input_strategy["version"] != FORMAT_VERSION:
            errors.append(
                f"Version mismatch. Expected {FORMAT_VERSION}, got {input_strategy['version']}"
            )

    # make sure the main components of the file are there
    unknown_headers = [
        header
        for header in input_strategy.keys()
        if header not in KNOWN_HEADERS and header not in KNOWN_OPERATIONS
    ]

    if unknown_headers:
        errors.append(f"{unknown_headers=} found. Supported headers: {KNOWN_HEADERS}")

    if "selfcal" in input_strategy:
        round_keys = input_strategy["selfcal"].keys()

        if not all([isinstance(i, int) for i in round_keys]):
            errors.append("The keys into the self-calibration should be ints. ")

        for round_info in round_keys:
            for mode in input_strategy["selfcal"][round_info]:
                try:
                    options = get_options_from_strategy(
                        strategy=input_strategy,
                        operation="selfcal",
                        mode=mode,
                        round_info=round_info,
                    )
                    try:
                        _ = MODE_OPTIONS_MAPPING[mode](**options)
                    except TypeError as typeerror:
                        errors.append(
                            f"{mode=} mode in {round_info=} incorrectly formed. {typeerror} "
                        )
                except Exception as exception:
                    errors.append(f"{exception}")

    if "polarisation" in input_strategy:
        _supported_polarisations = tuple(POLARISATION_MAPPING.keys())
        polarisations = input_strategy["polarisation"].keys()
        for polarisation in polarisations:
            if polarisation not in _supported_polarisations:
                errors.append(f"{polarisation=} not in {_supported_polarisations}. ")
            for mode in input_strategy["polarisation"][polarisation]:
                try:
                    options = get_options_from_strategy(
                        strategy=input_strategy,
                        operation="polarisation",
                        mode=mode,
                    )
                    try:
                        _ = MODE_OPTIONS_MAPPING[mode](**options)
                    except TypeError as typeerror:
                        errors.append(
                            f"{mode=} mode in {round_info=} incorrectly formed. {typeerror} "
                        )
                except Exception as exception:
                    errors.append(f"{exception}")

    for operation in KNOWN_OPERATIONS:
        # Already checked above
        if operation == "selfcal" or operation == "polarisation":
            continue
        if operation in input_strategy.keys():
            for mode in input_strategy[operation]:
                try:
                    options = get_options_from_strategy(
                        strategy=input_strategy, mode=mode, operation=operation
                    )
                    try:
                        _ = MODE_OPTIONS_MAPPING[mode](**options)
                    except TypeError as typeerror:
                        errors.append(
                            f"{mode=} mode in {operation=} incorrectly formed. {typeerror} "
                        )
                except Exception as exception:
                    errors.append(f"{exception}")

    valid_config = len(errors) == 0
    if not valid_config:
        for error in errors:
            logger.warning(error)

        if raise_on_error:
            raise ValueError("Configuration file not valid. ")

    return valid_config


def load_strategy_yaml(input_yaml: Path, verify: bool = True) -> Strategy:
    """Load in a flint based configuration file, which
    will be used to form the strategy for imaging of
    a field.

    The format of the return is likely to change. This
    is not to be relied on for the moment, and should
    be considered a toy. There will be a mutiny.

    Args:
        input_yaml (Path): The imaging strategy to use
        verify (bool, optional): Apply some basic checks to ensure a correctly formed strategy. Defaults to True.

    Returns:
        Strategy: The parameters of the imaging and self-calibration to use.
    """

    logger.info(f"Loading {input_yaml} file. ")

    with open(input_yaml) as in_file:
        input_strategy = Strategy(yaml.load(in_file, Loader=yaml.Loader))

    if verify:
        verify_configuration(input_strategy=input_strategy)

    return input_strategy


def write_strategy_to_yaml(strategy: Strategy, output_path: Path) -> Path:
    """Write the contents of a current strategy to a yaml file

    Args:
        strategy (Strategy): The strategy to write out
        output_path (Path): Where to write the output YAML file to

    Returns:
        Path: The path the output YAML file was written to
    """

    logger.info(f"Writing strategy to {output_path=}")

    with open(output_path, "w") as out_file:
        yaml.dump(data=strategy, stream=out_file, sort_keys=False)

    return output_path


# TODO: Create the file only for a subset of known defaults
def create_default_yaml(output_yaml: Path, selfcal_rounds: int | None = None) -> Path:
    """Create an example strategy yaml file that outlines the options to use at varies stages
    of some assumed processing pipeline.

    This is is completely experimental, and expected fields might change.

    Args:
        output_yaml (Path): Location to write the yaml file to.
        selfcal_rounds (Optional[int], optional): Will specify the number of self-calibration loops to include the file. If None, there will be none written. Defaults to None.

    Returns:
        Path: Path to the written yaml output file.
    """
    logger.info("Generating a default strategy. ")
    strategy: dict[Any, Any] = {}

    strategy["version"] = FORMAT_VERSION

    strategy["defaults"] = _create_mode_mapping_defaults()

    if selfcal_rounds:
        logger.info(f"Creating {selfcal_rounds} self-calibration rounds. ")
        selfcal: dict[int, Any] = {}
        for selfcal_round in range(0, selfcal_rounds):
            selfcal[selfcal_round] = {
                "wsclean": {},
                "gaincal": {},
                "masking": {},
            }

        strategy["selfcal"] = selfcal

    with open(output_yaml, "w") as out_file:
        logger.info(f"Writing {output_yaml}.")
        yaml.dump(data=strategy, stream=out_file, sort_keys=False)

    return output_yaml


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Tooling to interact with flint yaml configuration files. "
    )

    subparser = parser.add_subparsers(dest="mode")

    create_parser = subparser.add_parser(
        "create", help="Create an initial yaml file for editing. "
    )
    create_parser.add_argument(
        "output_yaml",
        type=Path,
        default="flint_strategy.yaml",
        help="The output YAML file to write with default options for various stages. ",
    )
    create_parser.add_argument(
        "--selfcal-rounds",
        type=int,
        default=None,
        help="Number of self-calibration rounds to use. ",
    )

    load_parser = subparser.add_parser("load")
    load_parser.add_argument(
        "input_yaml",
        type=Path,
        help="Path to a strategy yaml file to load and inspect. ",
    )
    verify_parser = subparser.add_parser(
        "verify", help="Verify a yaml file is correct, as far as we can tell.  "
    )
    verify_parser.add_argument("input_yaml", type=Path, help="Path to a strategy file")

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "create":
        create_default_yaml(
            output_yaml=args.output_yaml, selfcal_rounds=args.selfcal_rounds
        )
    elif args.mode == "load":
        load_strategy_yaml(input_yaml=args.input_yaml)
    elif args.mode == "verify":
        input_strategy = load_strategy_yaml(input_yaml=args.input_yaml)
        try:
            if verify_configuration(input_strategy=input_strategy):
                logger.info(f"{args.input_yaml} appears valid")
        except ValueError:
            logger.info(f"{args.input_yaml} does not appear to be valid")
    else:
        logger.error(f"{args.mode=} is not set or not known. Check --help. ")


if __name__ == "__main__":
    cli()



"""Helper utility functions to download and otherwise manage
containers required for flint"""

from __future__ import annotations

import asyncio
from argparse import ArgumentParser
from pathlib import Path

from flint.logging import logger
from flint.options import BaseOptions
from flint.sclient import pull_container


class FlintContainer(BaseOptions):
    """Item representing a Flint container"""

    name: str
    """Name of the container"""
    uri: str
    """URL of the container that can be used to pull with apptainer, e.g. docker://alecthomson/aoflagger:latest"""
    file_name: str
    """The expected file name of the container. This will be appended to the container directory path."""
    description: str | None = None
    """Short description on the purpose of the container"""
    tag: str = "v1.0"
    """Tag of the container, if applicable"""


calibrate_container = FlintContainer(
    name="calibrate",
    file_name="flint-containers_calibrate.sif",
    uri="docker://ghcr.io/flint-crew/calibrate",
    description="Contains AO calibrate and addmodel",
)
wsclean_container = FlintContainer(
    name="wsclean",
    file_name="flint-containers_wsclean.sif",
    uri="docker://ghcr.io/flint-crew/wsclean",
    description="Container with the wsclean deconvolution software",
)
askapsoft_contaer = FlintContainer(
    name="askapsoft",
    file_name="flint-containers_askapsoft.sif",
    uri="docker://ghcr.io/flint-crew/askapsoft",
    description="Container with askapsoft (also known as yandasoft)",
)
aoflagger_contaer = FlintContainer(
    name="aoflagger",
    file_name="flint-containers_aoflagger.sif",
    uri="docker://ghcr.io/flint-crew/aoflagger",
    description="Container with aoflagger, used to autonomously flag measurement sets",
)
aegean_contaer = FlintContainer(
    name="aegean",
    file_name="flint-containers_aegean.sif",
    uri="docker://ghcr.io/flint-crew/aegean",
    description="Container with aegean, used to source find",
)
potato_container = FlintContainer(
    name="potato",
    file_name="flint-containers_potato.sif",
    uri="docker://ghcr.io/flint-crew/potato",
    description="Peel out that terrible annoying source",
)
casa_container = FlintContainer(
    name="casa",
    file_name="flint-containers_casa.sif",
    uri="docker://ghcr.io/flint-crew/casa",
    description="A monolithic install of the CASA application",
)
LIST_OF_KNOWN_CONTAINERS = (
    calibrate_container,
    wsclean_container,
    askapsoft_contaer,
    aoflagger_contaer,
    aegean_contaer,
    potato_container,
    casa_container,
)
KNOWN_CONTAINER_LOOKUP: dict[str, FlintContainer] = {
    v.name: v for v in LIST_OF_KNOWN_CONTAINERS
}


def _sanity_check_containers(
    container_list: list[FlintContainer] | tuple[FlintContainer, ...],
) -> None:
    """Do some quick validation checks on the set of loaded containers. Make sure there are no
    duplicated names or file names"""
    assert len(container_list) == len(
        set([container.name for container in container_list])
    ), "Duplicated name detected in reference containers"
    assert len(container_list) == len(
        set([container.file_name for container in container_list])
    ), "Duplicated file name detected in reference containers"


_sanity_check_containers(container_list=LIST_OF_KNOWN_CONTAINERS)


def log_known_containers() -> None:
    """Log the known containers. This simply prints the set of known containers."""

    for idx, known_container in enumerate(LIST_OF_KNOWN_CONTAINERS):
        logger.info(f"Container {idx + 1} of {len(LIST_OF_KNOWN_CONTAINERS)}")
        logger.info(f"  Name: {known_container.name}")
        logger.info(f"  File name: {known_container.file_name}")
        logger.info(f"  URL: {known_container.uri}")
        logger.info(f"  Description: {known_container.description}")


def get_known_container_path(container_directory: Path | str, name: str) -> Path:
    """Return the path to a ``flint`` known container. These are containers that
    are downloaded through the ``download_known_containers`` function.

    Args:
        container_directory (Path | str): Path to directory containing downloaded containers
        name (str): Name of the container. Note that this is not the file name.

    Raises:
        ValueError: Raised when the name is not known

    Returns:
        Path: Path to the requested
    """

    container_directory = Path(container_directory)

    known_container = KNOWN_CONTAINER_LOOKUP.get(name, None)

    if known_container is None:
        raise ValueError(
            f"{name=} is not known. See {list(KNOWN_CONTAINER_LOOKUP.keys())}"
        )

    known_container_path = container_directory / known_container.file_name
    assert known_container_path.exists(), (
        f"{known_container_path=} of {name=} does not exist"
    )

    return known_container_path


def verify_known_containers(container_directory: Path | str) -> bool:
    """Inspect the provided ``container_directory`` to examine that the set of
    containers that are expected to exist are present.

    Args:
        container_directory (Path | str): Directory to search that should have containers

    Returns:
        bool: True is all containers are available. False otherwise.
    """

    logger.info(f"Checking {container_directory=} for known containers")
    container_valid = {}
    for known_container in LIST_OF_KNOWN_CONTAINERS:
        try:
            _ = get_known_container_path(
                container_directory=container_directory, name=known_container.name
            )
            valid = True
        except (ValueError, AssertionError):
            valid = False

        logger.info(
            f"Container {known_container.name} is {'valid' if valid else 'not valid'}"
        )
        container_valid[known_container.name] = valid

    return all(container_valid.values())


async def _pull_and_check_container(
    container_directory: Path,
    known_container: FlintContainer,
    expected_output_path: Path,
) -> None:
    """Pull a container and check that it was downloaded correctly

    Args:
        container_directory (Path): Output directory to store containers. Will be created if necessary.
        known_container (FlintContainer): Container to download
        expected_output_path (Path): Expected output path
    """
    _container_path = await asyncio.to_thread(
        pull_container,
        container_directory=container_directory,
        uri=f"{known_container.uri}:{known_container.tag}",
        file_name=known_container.file_name,
    )
    if not expected_output_path.exists():
        logger.error(
            f"{expected_output_path=} but was not. Instead received {_container_path=}"
        )


async def download_known_containers_coro(
    container_directory: Path | str, new_tag: str | None = None
) -> tuple[Path, ...]:
    """Download known containers for use throughout flint.

    Args:
        container_directory (Path | str): Output directory to store containers. Will be created if necessary.

    Returns:
        tuple[Path, ...]: Paths to all containers downloaded
    """
    container_directory = Path(container_directory)

    if not container_directory.exists():
        logger.info(f"Creating {container_directory=}")
        container_directory.mkdir(parents=True)

    containers_downloaded = []
    coros = []
    for idx, known_container in enumerate(LIST_OF_KNOWN_CONTAINERS):
        if new_tag is not None:
            known_container = known_container.with_options(tag=new_tag)
        logger.info(
            f"Downloading {idx + 1} of {len(LIST_OF_KNOWN_CONTAINERS)}, container {known_container.name}, tag {known_container.tag}"
        )

        expected_output_path = container_directory / known_container.file_name

        if expected_output_path.exists():
            logger.info(f"{expected_output_path=} already exists. Skipping.")
            continue

        coro = _pull_and_check_container(
            container_directory=container_directory,
            known_container=known_container,
            expected_output_path=expected_output_path,
        )
        coros.append(coro)

        containers_downloaded.append(expected_output_path)

    await asyncio.gather(*coros)

    logger.info(f"Downloaded {len(containers_downloaded)} new containers")
    return tuple(containers_downloaded)


def get_parser() -> ArgumentParser:
    """Create the CLI argument parser

    Returns:
        ArgumentParser: Constructed argument parser
    """
    parser = ArgumentParser(description="Download and manage Flint containers")
    subparsers = parser.add_subparsers(dest="mode")

    _ = subparsers.add_parser(
        name="list", help="List the containers that are known to Flint"
    )

    download_parser = subparsers.add_parser(
        name="download", help="Pull each of the known containers"
    )

    download_parser.add_argument(
        "container_directory", type=Path, help="Location to download containers to"
    )
    download_parser.add_argument(
        "--tag", type=str, help="Tag of the containers to download", default=None
    )

    verify_parser = subparsers.add_parser(
        name="verify", help="Pull each of the known containers"
    )

    verify_parser.add_argument(
        "container_directory", type=Path, help="Location to download containers to"
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "list":
        log_known_containers()
    elif args.mode == "download":
        asyncio.run(
            download_known_containers_coro(
                container_directory=args.container_directory, new_tag=args.tag
            )
        )
    elif args.mode == "verify":
        verify_known_containers(container_directory=args.container_directory)
    else:
        logger.info(f"Unknown directive: {args.mode}")


if __name__ == "__main__":
    cli()



"""Tests that are specific to the bandpass calibration
flow"""

from __future__ import annotations

from pathlib import Path

from flint.options import BandpassOptions, create_options_from_parser
from flint.prefect.flows import bandpass_pipeline


def test_bandpass_cli():
    """Ensure that the bandpass calibration using the BaseOptions
    class can integrate with the current preferred workflow
    used by racs-low3"""
    parser = bandpass_pipeline.get_parser()

    example_cli = """
    /some/test/argument
    --flagger-container /jack/sparrow/containers/aoflagger.sif
    --calibrate-container /jack/sparrow/containers/calibrate.sif
    --cluster-config ./petrichor.yaml
    --split-path /another/made/up/path
    --flag-calibrate-rounds 4
    --minuv 600
    --preflagger-jones-max-amplitude 0.6
    --preflagger-ant-mean-tolerance 0.18
    """
    args = parser.parse_args(example_cli.split())

    bandpass_options = create_options_from_parser(
        parser_namespace=args, options_class=BandpassOptions
    )
    assert args.bandpass_path == Path("/some/test/argument")
    assert args.split_path == Path("/another/made/up/path")
    assert args.cluster_config == "./petrichor.yaml"
    assert bandpass_options.flagger_container == Path(
        "/jack/sparrow/containers/aoflagger.sif"
    )
    assert bandpass_options.calibrate_container == Path(
        "/jack/sparrow/containers/calibrate.sif"
    )
    assert bandpass_options.minuv == 600.0
    assert isinstance(bandpass_options.minuv, float)
    assert not bandpass_options.preflagger_mesh_ant_flags
    assert bandpass_options.flag_calibrate_rounds == 4
    assert isinstance(bandpass_options.flag_calibrate_rounds, int)



"""Some specific tests around the pydantic base options model
that we are using to construct a BaseOptions class"""

from __future__ import annotations

from argparse import ArgumentParser

import pytest
from pydantic import ValidationError

from flint.options import BaseOptions, add_options_to_parser, create_options_from_parser


def test_ensure_options_frozen():
    """BaseOption classes should be immutable, so an error
    should be raised"""

    class NewOptions(BaseOptions):
        a: int
        """An example"""
        b: float
        """Another example"""

    new_options = NewOptions(a=1, b=1.23)
    with pytest.raises(ValidationError):
        # can't update the immutable class
        new_options.a = 33
        # raise error on argument not existing
        _ = NewOptions(a=1, b=1, jack="sparrow")  # type: ignore


def test_baseoptions_argparse():
    """Create an argument parser from a BaseOptions"""

    class NewOptions(BaseOptions):
        a: int
        """An example"""
        b: float
        """Another example"""
        c: bool = False
        """A flag"""

    parser = ArgumentParser(description="Jack Sparrow")

    parser = add_options_to_parser(parser=parser, options_class=NewOptions)
    args = parser.parse_args("1 1.23 --c".split())
    assert args.a == "1"
    assert isinstance(args.a, str)
    assert args.b == "1.23"
    assert isinstance(args.b, str)
    assert args.c
    assert isinstance(args.c, bool)

    new_options = create_options_from_parser(
        parser_namespace=args, options_class=NewOptions
    )
    assert isinstance(new_options, NewOptions)
    assert new_options.a == 1
    assert isinstance(new_options.a, int)
    assert new_options.b == 1.23
    assert isinstance(new_options.b, float)


def test_create_new_options():
    """Create a new subclass of BaseOptions"""

    class NewOptions(BaseOptions):
        a: int
        """An example"""
        b: float
        """Another example"""

    new_options = NewOptions(a=1, b=1.23)
    assert new_options.a == 1
    assert new_options.b == 1.23

    update_options = new_options.with_options(b=234.3)
    assert update_options.b == 234.3

    assert new_options is not update_options

    # Make sure the types are properly cast
    new_options = NewOptions(a=1, b=1)
    assert isinstance(new_options.b, float)


def test_create_new_options_asdict():
    """Create a new subclass of BaseOptions"""

    class NewOptions(BaseOptions):
        a: int
        """An example"""
        b: float
        """Another example"""

    new_options = NewOptions(a=1, b=1.23)
    _dict = new_options._asdict()
    assert isinstance(_dict, dict)
    assert _dict["a"] == 1
    assert isinstance(_dict["a"], int)
